{=========================================================================
  RkDBSpin Unit

  This unit implements the TRkDBSpinEdit component which is a data-aware
  version of the TSpinEdit Sample component provided with Delphi.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

{$I DCDC.INC}

unit RkDBSpin;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  {$IFDEF VER100}
  DBCtrls,
  {$ELSE}
  DBTables,
  {$ENDIF}
  DB, Forms, Spin, StdCtrls, RkCommon;

type
  EInvalidFieldType = class( Exception );

  TRkDBSpinEdit = class( TSpinEdit )
  private
    FAboutInfo : TRkAboutInfo;
    FAlignment : TAlignment;
    FDataLink : TFieldDataLink;
    FFocused : Boolean;
    FDisplayText : string;
    FCanvas : TControlCanvas;

    function GetDataField : string;
    procedure SetDataField( const Value : string );
    function GetDataSource : TDataSource;
    procedure SetDataSource( Value : TDataSource );
    function GetReadOnly : Boolean;
    procedure SetReadOnly( Value : Boolean );
    procedure CheckFieldType( const Value : string );

    procedure DataChange( Sender : TObject );
    procedure EditingChange( Sender : TObject );
    procedure UpdateData( Sender : TObject );
    procedure ActiveChange( Sender : TObject );

    procedure WMCut( var Msg : TMessage ); message wm_Cut;
    procedure WMPaste( var Msg : TMessage ); message wm_Paste;
    procedure WMPaint( var Msg : TWMPaint ); message wm_Paint;
    procedure UpdateFocus( Value : Boolean );
    procedure CMEnter( var Msg : TCMEnter ); message cm_Enter;
    procedure CMExit( var Msg : TCMExit ); message cm_Exit;
    procedure CMFontChanged( var Msg : TMessage ); message cm_FontChanged;
  protected
    procedure Notification( AComponent : TComponent;
                            Operation : TOperation ); override;
    procedure Change; override;
    procedure KeyDown( var Key : Word; Shift : TShiftState ); override;
    procedure KeyPress( var Key : Char ); override;
    procedure UpClick( Sender : TObject ); override;
    procedure DownClick( Sender : TObject ); override;
  public
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;
  published
    property About : TRkAboutInfo
      read FAboutInfo
      write FAboutInfo
      stored False;

    property DataField : string
      read GetDataField
      write SetDataField;

    property DataSource : TDataSource
      read GetDataSource
      write SetDataSource;

    // This property controls the ReadOnly state of the DataLink
    property ReadOnly : Boolean
      read GetReadOnly
      write SetReadOnly
      default False;

    { Inherited Properties }
    property BorderStyle;
    property Width default 50;
  end;


implementation


{===========================}
{== TRkDBSpinEdit Methods ==}
{===========================}

constructor TRkDBSpinEdit.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );

  // Cannot replicate because button controls created by TSpinEdit
  // are not replicatable.
  ControlStyle := ControlStyle - [ csReplicatable ];

  Width := 50;

  inherited ReadOnly := True;     // Sets the edit control to be read-only

  FDataLink := TFieldDataLink.Create;

  // To support the TField.FocusControl method, set the FDataLink.Control
  // property to point to the spin button.  The Control property requires
  // a TWinControl component.
  FDataLink.Control := Self;

  FDataLink.OnDataChange := DataChange;           // Assign event handlers
  FDataLink.OnUpdateData := UpdateData;
  FDataLink.OnEditingChange := EditingChange;
  FDataLink.OnActiveChange := ActiveChange;
end;


destructor TRkDBSpinEdit.Destroy;
begin
  FDataLink.Free;
  FDataLink := nil;
  FCanvas.Free;
  inherited Destroy;
end;


procedure TRkDBSpinEdit.Notification( AComponent : TComponent;
                                      Operation : TOperation );
begin
  inherited Notification( AComponent, Operation );
  if ( Operation = opRemove ) and
     ( FDataLink <> nil ) and
     ( AComponent = DataSource ) then
    DataSource := nil;
end;


function TRkDBSpinEdit.GetDataField : string;
begin
  Result := FDataLink.FieldName;
end;


procedure TRkDBSpinEdit.SetDataField( const Value : string );
begin
  CheckFieldType( Value );             // Check that field type is Integer
  FDataLink.FieldName := Value;
end;


function TRkDBSpinEdit.GetDataSource : TDataSource;
begin
  Result := FDataLink.DataSource;
end;


procedure TRkDBSpinEdit.SetDataSource( Value : TDataSource );
begin
  if FDatalink.DataSource <> Value then
  begin
    FDataLink.DataSource := Value;

    if Value <> nil then
      Value.FreeNotification( Self );
  end;
end;


function TRkDBSpinEdit.GetReadOnly : Boolean;
begin
  Result := FDataLink.ReadOnly;
end;


procedure TRkDBSpinEdit.SetReadOnly( Value : Boolean );
begin
  FDataLink.ReadOnly := Value;
end;


{=========================================================================
  TRkDBSpinEdit.CheckFieldType

  This method checks to make sure the field type corresponding to the
  column referenced by Value is either ftInteger, ftSmallInt, or ftWord.
  If it is not, then an EInvalidFieldType exception is raised.
=========================================================================}

procedure TRkDBSpinEdit.CheckFieldType( const Value : string );
var
  FieldType : TFieldType;
begin
  if ( Value <> '' ) and
     ( FDataLink <> nil ) and
     ( FDataLink.Dataset <> nil ) and
     ( FDataLink.Dataset.Active ) then
  begin
    FieldType := FDataLink.Dataset.FieldByName( Value ).DataType;
    if ( FieldType <> ftInteger ) and
       ( FieldType <> ftSmallInt ) and
       ( FieldType <> ftWord ) then
    begin
      raise EInvalidFieldType.Create( 'RkDBSpinEdit.DataField can only ' +
                                      'be connected to columns of type ' +
                                      'Integer' );
    end;
  end;
end;


procedure TRkDBSpinEdit.Change;
begin
  if FDataLink <> nil then
    FDataLink.Modified;    // Tell the FDataLink that the data has changed
  inherited Change;
end;


procedure TRkDBSpinEdit.KeyDown( var Key : Word; Shift : TShiftState );
begin
  inherited KeyDown( Key, Shift );

  // Need to handle Old Fashioned Cut, Paste, and Delete keystrokes
  if ( Key = vk_Delete ) or
     ( ( Key = vk_Insert ) and ( ssShift in Shift ) ) then
    FDataLink.Edit;
end;


procedure TRkDBSpinEdit.KeyPress( var Key : Char );
begin
  inherited KeyPress( Key );

  // The field object referenced by FDataLink will tell us which keys
  // are valid

  if ( Key in [ #32..#255 ] ) and
     ( FDataLink.Field <> nil ) and
     not FDataLink.Field.IsValidChar( Key ) then
  begin
    MessageBeep( 0 );
    Key := #0;                               // Invalid keys are discarded
  end;

  case Key of
    ^H, ^V, ^X, #32..#255:
      FDataLink.Edit;          // Put corresponding Dataset into Edit mode

    #27:                                             // Escape key pressed
    begin
      FDataLink.Reset;
      SelectAll;
      Key := #0;
    end;
  end;
end;


procedure TRkDBSpinEdit.UpClick( Sender : TObject );
begin
  if ReadOnly then
    MessageBeep( 0 )            // Prevent change if FDataLink is ReadOnly
  else
  begin
    if FDataLink.Edit then     // Put corresponding Dataset into edit mode
      Value := Value + Increment;        // Increment only if in edit mode
  end;
end;


procedure TRkDBSpinEdit.DownClick( Sender : TObject );
begin
  if ReadOnly then
    MessageBeep( 0 )            // Prevent change if FDataLink is ReadOnly
  else
  begin
    if FDataLink.Edit then     // Put corresponding Dataset into edit mode
      Value := Value - Increment;        // Decrement only if in edit mode
  end;
end;


{=========================================================================
  TRkDBSpinEdit.DataChange

  This method gets called as a result of a number of different events:

  1. The underlying field value changes.  Occurs when changing the value
     of the column tied to this control and then move to a new column or
     a new record.
  2. The corresponding Dataset goes into Edit mode.
  3. The corresponding Dataset referenced by DataSource changes.
  4. The current cursor is scrolled to a new record in the table.
  5. The record is reset through a Cancel call.
  6. The DataField property changes to reference another column.
=========================================================================}

procedure TRkDBSpinEdit.DataChange( Sender : TObject );
begin
  if FDataLink.Field <> nil then
  begin
    if FAlignment <> FDataLink.Field.Alignment then
    begin
      Text := '';   // Forces the Text assignment below to cause a Repaint
      FAlignment := FDataLink.Field.Alignment;
    end;

    Text := FDataLink.Field.Text;
    if FFocused and FDataLink.CanModify then
      FDisplayText := Text
    else
      FDisplayText := FDataLink.Field.DisplayText;
  end
  else                                     // There is no field assignment
  begin
    FAlignment := taLeftJustify;
    if csDesigning in ComponentState then
      Text := Name               // In design-mode, show Name of component
    else
      Text := '';
  end;
end; {= TRkDBSpinEdit.DataChange =}


{=========================================================================
  TRkDBSpinEdit.UpdateData

  This method gets called when the corresponding field value and the
  contents of the SpinEdit need to be synchronized.  Note that this
  method only gets called if this control was responsible for altering
  the data.
=========================================================================}

procedure TRkDBSpinEdit.UpdateData(Sender: TObject);
begin
  // You can cancel the Update process by raising an exception
  // in this method.

  FDataLink.Field.Text := Text;
end;


{=========================================================================
  TRkDBSpinEdit.EditingChange
                                               
  This method gets called when the State of the attached Dataset (via the
  DataSource) changes between an Editing state and a Non-Editing state.
  The different states are described below:

  Editing States       Non-Editing States
  --------------       ------------------
  dsEdit               dsInactive
  dsInsert             dsBrowse
  dsSetKey             dsCalcFields

  NOTE: This method only gets called if the DataLink is not ReadOnly.
=========================================================================}

procedure TRkDBSpinEdit.EditingChange(Sender: TObject);
begin
  // The ReadOnly property of the underlying edit control is toggled
  // according to the state of the DataSource.

  inherited ReadOnly := not FDataLink.Editing;
end;


{=========================================================================
  TRkDBSpinEdit.ActiveChange

  This method gets called whenever the Active property of the attached
  Dataset changes.

  NOTE: You can use the FDataLink.Active property to determine the *new*
        state of the Dataset.
=========================================================================}

procedure TRkDBSpinEdit.ActiveChange( Sender : TObject );
begin
  // If the Dataset is becoming Active, then check to make sure the
  // field type of the DataField property is an Integer type.

  if ( FDataLink <> nil ) and FDataLink.Active then
    CheckFieldType( DataField );
end;


{=========================================================================
  Both the WMPaste and WMCut methods had to be copied from the TSpinEdit
  class because placing the Dataset into Edit mode must come before
  altering the text, but after checking for ReadOnly.
=========================================================================}

procedure TRkDBSpinEdit.WMPaste( var Msg : TMessage );
begin
  // EditorEnabled is defined in TSpinEdit
  if not EditorEnabled or ReadOnly then
    Exit;
  FDataLink.Edit;
  inherited;
end;


procedure TRkDBSpinEdit.WMCut( var Msg : TMessage );
begin
  if not EditorEnabled or ReadOnly then
    Exit;
  FDataLink.Edit;
  inherited;
end;


{=========================================================================
  TRkDBSpinEdit.WMPaint

  This method takes a trick provided in the DBCtrls unit for handling the
  display of right justified text into an edit field.  If the text needs
  to be drawn right-justified, then the text is drawn manually. Otherwise,
  the default message handler handles it.  The contents are always drawn
  left-justified when the control has the focus.
=========================================================================}

procedure TRkDBSpinEdit.WMPaint( var Msg : TWMPaint );
var
  Width, Margin, X : Integer;
  EditRct, TempRct : TRect;
  DC : HDC;
  PS : TPaintStruct;
begin
  if ( FAlignment = taLeftJustify ) or FFocused then
  begin
    inherited;
    Exit;
  end;

  if FCanvas = nil then
  begin
    // Need something to draw on, so create a control canvas
    FCanvas := TControlCanvas.Create;
    FCanvas.Control := Self;
  end;
  DC := Msg.DC;
  if DC = 0 then
    DC := BeginPaint( Handle, PS );          // Setup Windows for painting
  FCanvas.Handle := DC;                // Connect Canvas to device context

  try
    FCanvas.Font := Font;
    with FCanvas do
    begin
     // Because the SpinButton is inside the edit field, the entire client
     // area is not used for drawing text.  Send the em_GetRect method to
     // get the rectangle to use.  Adjust the right side to avoid drawing
     // under spin button

      EditRct := ClientRect;
      Perform( em_GetRect, 0, LongInt( @TempRct ) );
      EditRct.Right := TempRct.Right;

      Brush.Color := Color;
      Width := TextWidth( FDisplayText );

      if BorderStyle = bsNone then
        Margin := 0
      else
        Margin := 1;

      if FAlignment = taRightJustify then
        X := EditRct.Right - Width - Margin
      else                                      // Handle center alignment
        X := ( EditRct.Left + EditRct.Right - Width ) div 2;
      TextRect( EditRct, X, Margin, FDisplayText );       // Draw the text
    end;
  finally
    FCanvas.Handle := 0;
    if Msg.DC = 0 then
      EndPaint( Handle, PS );
  end;

end;


procedure TRkDBSpinEdit.UpdateFocus( Value : Boolean );
begin
  // Because Alignment automatically reverts back to taLeftJustify when
  // control gets the focus, the control must be repainted, when the
  // focus changes.

  if FFocused <> Value then
  begin
    FFocused := Value;
    if FAlignment <> taLeftJustify then
      Invalidate;
    FDataLink.Reset;
  end;
end;


procedure TRkDBSpinEdit.CMEnter( var Msg : TCMEnter );
begin
  UpdateFocus( True );
  inherited;
end;


procedure TRkDBSpinEdit.CMExit( var Msg : TCMExit );
begin
  try        // Attempt to Update the record if focus leaves the spin edit
    FDataLink.UpdateRecord;
  except
    SelectAll;
    SetFocus;             // Keep the focus on the control if Update fails
    raise;                                        // Reraise the exception
  end;

  UpdateFocus( False );
  inherited;
end;


procedure TRkDBSpinEdit.CMFontChanged( var Msg : TMessage );
begin
  inherited;
  SendMessage( Handle, wm_Size, 0, 0 );
end;


end.
