{=========================================================================
  RkNavPnl Unit

  This unit implements the TRkNavigatorPanel component, which demonstrates
  how to add a drop area to a compound component. A drop area is an area,
  typically a panel or scrollbox, within the compound component that can
  accept other controls dropped onto it.  However, the drop area itself
  cannot be selected using the Object Inspector.

  The TRkNavigatorPanel component embeds a DBNavigator and a ScrollBox
  within a panel.  The ScrollBox serves as the drop area.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

{$I DCDC.INC}

unit RkNavPnl;

interface

uses
  Windows, SysUtils, Classes, Controls, ExtCtrls, Forms, DB, DBCtrls,
  RkCommon;

type
  TRkNavigatorPanel = class( TPanel )
  private
    FAboutInfo : TRkAboutInfo;
    // Embedded ScrollBox will accept other components dropped onto it
    FDropArea : TScrollBox;
    // Declare an internal string list for each drop area component
    FDropAreaList : TStringList;

    // Other embedded components
    FNavigator : TDBNavigator;
    FSpacer : TPanel;
    function GetDataSource : TDataSource;
    procedure SetDataSource( Value : TDataSource );
    function GetDropAreaBorder : TBorderStyle;
    procedure SetDropAreaBorder( Value : TBorderStyle );
    {$IFDEF VER100}
    // Flat style buttons are only available in Delphi 3.0
    function GetFlatButtons : Boolean;
    procedure SetFlatButtons( Value : Boolean );
    {$ENDIF}
    procedure CreateNavigator;
    procedure CreateSpacer;
    procedure CreateDropArea;
  protected
    // These five protected methods must be overridden to ensure that any
    // components dropped onto the drop area are correctly streamed to
    // the form file.
    procedure Loaded; override;
    {$IFDEF VER100}
    procedure GetChildren( Proc : TGetChildProc;
                           Root : TComponent ); override;
    {$ELSE}
    procedure GetChildren( Proc : TGetChildProc ); override;
    {$ENDIF}
    procedure DefineProperties( Filer : TFiler ); override;
    procedure ReadDropList( Reader : TReader );
    procedure WriteDropList( Writer : TWriter );
  public
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;

    property DropArea : TScrollBox         // Run-time access to drop area
      read FDropArea;
  published
    property About : TRkAboutInfo
      read FAboutInfo
      write FAboutInfo
      stored False;

    // Exposing properties of embedded components

    property DropAreaBorder : TBorderStyle
      read GetDropAreaBorder
      write SetDropAreaBorder;

    property DataSource : TDataSource
      read GetDataSource
      write SetDataSource;

    {$IFDEF VER100}
    property FlatButtons : Boolean
      read GetFlatButtons
      write SetFlatButtons
      default False;
    {$ENDIF}

    property BevelOuter default bvNone;
    property BorderStyle default bsSingle;
  end;


implementation


{===============================}
{== TRkNavigatorPanel Methods ==}
{===============================}

constructor TRkNavigatorPanel.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );
  ControlStyle := ControlStyle - [ csAcceptsControls, csSetCaption ];
  Width := 250;
  Height := 150;
  BorderWidth := 4;

  CreateNavigator;
  CreateSpacer;
  CreateDropArea;
end;


procedure TRkNavigatorPanel.CreateNavigator;
begin
  FNavigator := TDBNavigator.Create( Self );
  FNavigator.Parent := Self;
  FNavigator.Align := alTop;
  FNavigator.ShowHint := True;
  FNavigator.Enabled := False;
  FNavigator.Enabled := True;     // Ensure nav buttons show correct state
end;


procedure TRkNavigatorPanel.CreateSpacer;
begin
  FSpacer := TPanel.Create( Self );
  with FSpacer do
  begin
    Parent := Self;
    // Do not allow components to be dropped onto the spacer
    ControlStyle := ControlStyle - [ csAcceptsControls ];
    TabStop := False;
    Align := alTop;
    Height := 4;
    BevelOuter := bvNone;
  end;
end;


{=========================================================================
  TRkNavigatorPanel.CreateDropArea

  This method is responsible for creating the FDropArea instance. This
  method also creates the string list to be associated with the drop area.
  The string list will be used to store which components are contained
  within the drop area when the component is written to a stream.
=========================================================================}

procedure TRkNavigatorPanel.CreateDropArea;
begin
  FDropArea := TScrollBox.Create( Self );
  FDropArea.Parent := Self;
  FDropArea.Align := alClient;
  FDropArea.TabStop := False;

  FDropAreaList := TStringList.Create;        // Create the drop area list
end;


destructor TRkNavigatorPanel.Destroy;
begin
  FDropAreaList.Free;           // Don't forget to destroy the string list
  inherited Destroy;
end;


{=========================================================================
  TRkNavigatorPanel.Loaded

  This method is responsible for "correcting" the parent hierarchy within
  the compound component. That is, when the component is written to a
  stream, the Parent of all controls dropped onto the drop area are
  changed to the main component rather than the drop area. This is
  necessary in order to support form inheritance.  This method, corrects
  the Parent hierarchy by iterating through the FDropAreaList.
=========================================================================}

procedure TRkNavigatorPanel.Loaded;
var
  I : Integer;
  C : TComponent;
begin
  inherited Loaded;

  for I := 0 to FDropAreaList.Count - 1 do
  begin
    C := GetParentForm( Self ).FindComponent( FDropAreaList[ I ] );
    if C <> nil then
      TControl( C ).Parent := FDropArea;
  end;
end;


{=========================================================================
  TRkNavigatorPanel.GetChildren

  This method is responsible for changing the parent hierarchy when the
  component is written to a stream. This method is called by Delphi to
  determine which components are children of the main component. We
  override this method to tell Delphi to treat all of the components
  dropped into the drop area as children of the main component rather
  than children of the embedded ScrollBox. The parent hierarchy is
  corrected in the Loaded method.
=========================================================================}

{$IFDEF VER100}
procedure TRkNavigatorPanel.GetChildren( Proc : TGetChildProc;
                                         Root : TComponent );
{$ELSE}
procedure TRkNavigatorPanel.GetChildren( Proc : TGetChildProc );
{$ENDIF}
var
  I : Integer;
  Control : TControl;
  {$IFDEF VER100}
  Form : TCustomForm;
  {$ELSE}
  Form : TForm;
  {$ENDIF}
begin
  {$IFDEF VER100}
  inherited GetChildren( Proc, Root );
  {$ELSE}
  inherited GetChildren( Proc );
  {$ENDIF}

  Form := GetParentForm( Self );
  for I := 0 to FDropArea.ControlCount - 1 do
  begin                               // For each control in the drop area
    Control := FDropArea.Controls[ I ];
    // Only call Proc if the control's Owner is the current form. This
    // will not be true when form inheritance is being used.
    if Control.Owner = Form then
      Proc( Control );
  end;
end;


{=========================================================================
  TRkNavigatorPanel.DefineProperties

  In order to correct the parent hierarchy, a list of all controls that
  were dropped into the drop area must be stored in the stream along
  with the component. To do this, we define a new custom property using
  this method.
=========================================================================}

procedure TRkNavigatorPanel.DefineProperties( Filer : TFiler );
begin
  Filer.DefineProperty('DropAreaList', ReadDropList, WriteDropList, True);
end;


{=========================================================================
  TRkNavigatorPanel.ReadDropList

  This method is called whenever the DropAreaList custom property needs
  to be read from a stream.  Each component identifier that was stored in
  the stream is read and saved into the FDropAreaList string list.
=========================================================================}

procedure TRkNavigatorPanel.ReadDropList( Reader : TReader );
begin
  FDropAreaList.Clear;
  Reader.ReadListBegin;
  while not Reader.EndOfList do
    FDropAreaList.Add( Reader.ReadIdent );
  Reader.ReadListEnd;
end;


{=========================================================================
  TRkNavigatorPanel.WriteDropList

  This method is called whenever the DropAreaList custom property needs
  to be written to a stream.  The identifier of each control dropped onto
  the drop area is stored in the stream.
=========================================================================}

procedure TRkNavigatorPanel.WriteDropList( Writer : TWriter );
var
  I : Integer;
begin
  Writer.WriteListBegin;
  for I := 0 to FDropArea.ControlCount - 1 do
    Writer.WriteIdent( FDropArea.Controls[ I ].Name );
  Writer.WriteListEnd;
end;


function TRkNavigatorPanel.GetDropAreaBorder : TBorderStyle;
begin
  Result := FDropArea.BorderStyle;
end;

procedure TRkNavigatorPanel.SetDropAreaBorder( Value : TBorderStyle );
begin
  FDropArea.BorderStyle := Value;
end;


function TRkNavigatorPanel.GetDataSource : TDataSource;
begin
  Result := FNavigator.DataSource;
end;

procedure TRkNavigatorPanel.SetDataSource( Value : TDataSource );
begin
  FNavigator.DataSource := Value;
end;


{$IFDEF VER100}
function TRkNavigatorPanel.GetFlatButtons : Boolean;
begin
  Result := FNavigator.Flat;
end;

procedure TRkNavigatorPanel.SetFlatButtons( Value : Boolean );
begin
  FNavigator.Flat := Value;
end;
{$ENDIF}


end.
