{=========================================================================
  RkTabLst Unit

  This unit implements an enhanced list box component which understands
  how to process embedded tab characters to display lines in columnar
  format. In addition, a horizontal scroll bar is added to the list box.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

{$I DCDC.INC}

unit RkTabLst;

interface

uses
  Windows, Classes, Graphics, Forms, StdCtrls,
  Controls, Messages, SysUtils, IntList, RkCommon;

type
  TRkTabbedListBox = class;

  TRkTabStopList = class( TIntegerList )
  private
    FListBox : TRkTabbedListBox;
  protected
    procedure SetItem( Index : Integer; Value : Integer ); override;
  public
    constructor Create;
    procedure Delete( Index : Integer ); override;
    procedure Insert( Index : Integer; Value : Integer ); override;
  end;


  TRkTabbedListBox = class( TListBox )
  private
    FAboutInfo : TRkAboutInfo;
    FHorzExtent : Word;
    FHorzScrollBar : Boolean;
    FTabStops : TRkTabStopList;
    procedure SetHorzExtent( Value : Word );
    procedure SetHorzScrollBar( Value : Boolean );
    procedure SetTabStops( Value : TRkTabStopList );
  protected
    procedure CreateParams( var Params : TCreateParams ); override;
    procedure CreateWnd; override;
  public
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;
    procedure UpdateTabStops;
  published
    property About : TRkAboutInfo
      read FAboutInfo
      write FAboutInfo
      stored False;

    property HorzExtent : Word
      read FHorzExtent
      write SetHorzExtent
      default 0;

    property HorzScrollBar : Boolean
      read FHorzScrollBar
      write SetHorzScrollBar
      default False;

    property TabStops : TRkTabStopList               
      read FTabStops
      write SetTabStops;
  end;


implementation


{============================}
{== TRkTabStopList Methods ==}
{============================}

constructor TRkTabStopList.Create;
begin
  inherited Create;
  Min := 0;
  Max := MaxLongint;
end;


procedure TRkTabStopList.SetItem( Index : Integer; Value : Integer );
begin
  inherited SetItem( Index, Value );
  if FListBox <> nil then
    FListBox.UpdateTabStops;
end;


procedure TRkTabStopList.Delete( Index : Integer );
begin
  inherited Delete( Index );
  if FListBox <> nil then
    FListBox.UpdateTabStops;
end;


procedure TRkTabStopList.Insert( Index : Integer; Value : Integer );
begin
  inherited Insert( Index, Value );
  if FListBox <> nil then
    FListBox.UpdateTabStops;
end;



{==============================}
{== TRkTabbedListBox Methods ==}
{==============================}

constructor TRkTabbedListBox.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );
  FTabStops := TRkTabStopList.Create;
  FTabStops.FListBox := Self;
end;


destructor TRkTabbedListBox.Destroy;
begin
  FTabStops.Free;
  inherited Destroy;
end;


procedure TRkTabbedListBox.CreateParams( var Params : TCreateParams );
begin
  inherited CreateParams( Params );      // Don't forget to call inherited
  with Params do
    Style := Style or lbs_UseTabStops or ws_HScroll;
end;


procedure TRkTabbedListBox.CreateWnd;
begin
  inherited CreateWnd;                   // Don't forget to call inherited

  // Initializing the scroll bar must occur after the window handle for
  // the list box has been created

  Perform( lb_SetHorizontalExtent, FHorzExtent, 0 );
  ShowScrollBar( Handle, sb_Horz, FHorzScrollBar );
  UpdateTabStops;
end;


procedure TRkTabbedListBox.SetHorzExtent( Value : Word );
begin
  if FHorzExtent <> Value then
  begin
    FHorzExtent := Value;
    Perform( lb_SetHorizontalExtent, FHorzExtent, 0 );
  end;
end;


procedure TRkTabbedListBox.SetHorzScrollBar( Value : Boolean );
begin
  if FHorzScrollBar <> Value then
  begin
    FHorzScrollBar := Value;
    ShowScrollBar( Handle, sb_Horz, FHorzScrollBar );
  end;
end;


procedure TRkTabbedListBox.SetTabStops( Value : TRkTabStopList );
begin
  FTabStops.Assign( Value );
  UpdateTabStops;
end;


{=========================================================================
  TRkTabbedListBox.UpdateTabStops

  The lb_SetTabStops message is used to set the tab stops.  The LParam
  parameter for this message is a pointer to an array of Word values
  representing the tab stops.  Therefore, the contents of the FTabStops
  list is transferred to a temporary array.
=========================================================================}

procedure TRkTabbedListBox.UpdateTabStops;
const
  MaxTabs = 1000;                      // 1000 tab stops should be plenty!
var
  TabCount, I : Integer;
  TabArray : array[ 0..MaxTabs - 1 ] of Integer;
begin
  TabCount := FTabStops.Count;
  if TabCount > MaxTabs then
    TabCount := MaxTabs;

  // Copy contents of FTabStops to temp array and convert chars to
  // dialog units

  for I := 0 to TabCount - 1 do
    TabArray[ I ] := FTabStops.Items[ I ] * 4;

                              // Send message to list box to set tab stops
  Perform( lb_SetTabStops, TabCount, Longint( @TabArray ) );
  Invalidate;
end;


end.


